#pragma once

#include <cstdint>
#include <fstream>
#include <functional>

#include <amdis/Assembler.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/typetree/Traversal.hpp>
#include <amdis/utility/AssembleOperators.hpp>

namespace AMDiS {

template <class GB, class B>
void DOFVectorBase<GB,B>::
init(bool asmVector)
{
  backend_.resize(size());
  if (asmVector)
    backend_.set(0);

  auto const localSize = basis_->localView().maxSize();
  elementVector_.resize(localSize);
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
finish(bool /*asmVector*/)
{}


template <class GB, class B>
void DOFVectorBase<GB,B>::
insert(LocalView const& localView, ElementVector const& elementVector)
{
  using std::abs;
  for (size_type i = 0; i < localView.size(); ++i) {
    if (abs(elementVector[i]) > threshold<double>) {
      size_type const idx = flatMultiIndex(localView.index(i));
      backend_[idx] += elementVector[i];
    }
  }
}


template <class GB, class B>
  template <class ContextTag, class Expr, class TreePath>
void DOFVectorBase<GB,B>::
addOperator(ContextTag contextTag, Expr const& expr, TreePath path)
{
  static_assert( Concepts::PreTreePath<TreePath>,
      "path must be a valid treepath, or an integer/index-constant");

  auto i = child(basis_->localView().tree(), makeTreePath(path));

  using LocalContext = typename ContextTag::type;
  using Traits = DefaultAssemblerTraits<LocalContext, ElementVector>;
  auto op = makeLocalOperator<LocalContext>(expr, basis_->gridView());
  auto localAssembler = makeUniquePtr(makeAssembler<Traits>(std::move(op), i));

  operators_[i].push(contextTag, std::move(localAssembler));
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
assemble(LocalView const& localView)
{
  elementVector_ = 0;
  auto const& gv = basis_->gridView();
  auto const& element = localView.element();
  auto geometry = element.geometry();

  for_each_node(localView.tree(), [&](auto const& node, auto) {
    auto& rhsOp = operators_[node];
    if (rhsOp) {
      rhsOp.bind(element, geometry);
      assembleOperators(gv, element, rhsOp, makeVectorAssembler(node, elementVector_));
      rhsOp.unbind();
    }
  });

  insert(localView, elementVector_);
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
assemble()
{
  auto localView = basis_->localView();

  init(true);
  for (auto const& element : elements(basis_->gridView())) {
    localView.bind(element);
    assemble(localView);
    localView.unbind();
  }
  finish(true);
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
backup(std::string const& filename)
{
  std::ofstream out(filename, std::ios::binary);

  std::int64_t numElements = basis_->gridView().size(0);
  out.write((char*)&numElements, sizeof(std::int64_t));

  auto localView = basis_->localView();
  std::vector<value_type> data;
  for (auto const& element : elements(basis_->gridView()))
  {
    localView.bind(element);
    data.clear();
    for_each_leaf_node(localView.tree(), [&](auto const& node, auto) -> void {
      auto const& fe = node.finiteElement();
      std::size_t size = fe.size();

      for (std::size_t i = 0; i < size; ++i)
        data.push_back((*this)[localView.index(node.localIndex(i))]);
    });

    std::uint64_t len = data.size();
    out.write((char*)&len, sizeof(std::uint64_t));
    out.write((char*)data.data(), len*sizeof(value_type));

    localView.unbind();
  }
}

template <class GB, class B>
void DOFVectorBase<GB,B>::
restore(std::string const& filename)
{
  std::ifstream in(filename, std::ios::binary);

  std::int64_t numElements = 0;
  in.read((char*)&numElements, sizeof(std::int64_t));
  assert(numElements == basis_->gridView().size(0));

  // assume the order of element traversal is not changed
  auto localView = basis_->localView();
  std::vector<value_type> data;
  for (auto const& element : elements(basis_->gridView()))
  {
    std::uint64_t len = 0;
    in.read((char*)&len, sizeof(std::uint64_t));
    data.resize(len);

    in.read((char*)data.data(), len*sizeof(value_type));

    localView.bind(element);
    std::size_t shift = 0;
    for_each_leaf_node(localView.tree(), [&](auto const& node, auto) -> void {
      auto const& fe = node.finiteElement();
      std::size_t size = fe.size();

      assert(data.size() >= shift+size);
      for (std::size_t i = 0; i < size; ++i)
        (*this)[localView.index(node.localIndex(i))] = data[shift + i];

      shift += size;
    });

    localView.unbind();
  }
}

} // end namespace AMDiS
